/*********************************************************************
*                    SEGGER Microcontroller GmbH                     *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*            (c) 2014 - 2020 SEGGER Microcontroller GmbH             *
*                                                                    *
*           www.segger.com     Support: support@segger.com           *
*                                                                    *
**********************************************************************
*                                                                    *
* All rights reserved.                                               *
*                                                                    *
* Redistribution and use in source and binary forms, with or         *
* without modification, are permitted provided that the following    *
* conditions are met:                                                *
*                                                                    *
* - Redistributions of source code must retain the above copyright   *
*   notice, this list of conditions and the following disclaimer.    *
*                                                                    *
* - Neither the name of SEGGER Microcontroller GmbH                  *
*   nor the names of its contributors may be used to endorse or      *
*   promote products derived from this software without specific     *
*   prior written permission.                                        *
*                                                                    *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND             *
* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,        *
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF           *
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE           *
* DISCLAIMED.                                                        *
* IN NO EVENT SHALL SEGGER Microcontroller GmbH BE LIABLE FOR        *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR           *
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT  *
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;    *
* OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF      *
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT          *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE  *
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH   *
* DAMAGE.                                                            *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

File    : main.c
Purpose : QSPI example application for the Renesas 1500KB EVK Board

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define U32 uint32_t
#define U16 uint16_t
#define U8  uint8_t
//
// QSPI controller registers
//
#define QSPI_CONTROLLER_BASE_ADDR (0x64000000)
#define QSPI                      ((_QSPIController*)QSPI_CONTROLLER_BASE_ADDR)
//
// Pin configuration unlock register
//
#define PMISC_PWPR                (*((volatile U8*) (0x40040D03)))
//
// Pin function config for the RTK70E015DC00001BJ board
//
#define PFS_QSPCLK_A              (*((volatile U32*)(0x40040988)))         // Pin used: P602
#define PFS_QSSL_A                (*((volatile U32*)(0x40040984)))         // Pin used: P601
#define PFS_QIO1_A                (*((volatile U32*)(0x4004083C)))         // Pin used: P015
#define PFS_QIO0_A                (*((volatile U32*)(0x40040940)))         // Pin used: P500
//                                
// Module stop state              
//                                
#define MSTP_MSTPCRB              (*((volatile U32*)(0x40047000)))
#define SYSTEM_VOCR               (*((volatile U8*) (0x4001E440)))
#define SYSTEM_PRCR               (*((volatile U16*)(0x4001E3FE)))
//                                
// Watchdog timers                
//                                
#define WDT_WDTRR                 (*((volatile U8*) (0x40044200)))
#define IWDT_IWDTRR               (*((volatile U8*) (0x40044400)))
                                  
#define FLASH_BASE_ADDR           (0x60000000)

/*********************************************************************
*
*       Types
*
**********************************************************************
*/

typedef struct {
  volatile U32 SFMSMD;
  volatile U32 SFMSSC;
  volatile U32 SFMSKC;
  volatile U32 SFMSST;
  volatile U32 SFMCOM;
  volatile U32 SFMCMD;
  volatile U32 SFMCST;
  volatile U32 ReservedReg1;
  volatile U32 SFMSIC;
  volatile U32 SFMSAC;
  volatile U32 SFMSDC;
  volatile U32 ReservedReg2;
  volatile U32 SFMSPC;
  volatile U32 SFMPMD;
} _QSPIController;

/*********************************************************************
*
*       Static code
*
**********************************************************************
*/

/*********************************************************************
*
*       _InitClocks()
*
*  Function description
*   Initiates the system clock.
*/
static void _InitClocks(void) {
  U16 rSYSTEM_PRCR;
  //
  // Here we just stick to the system clock that is selected after reset
  //
  rSYSTEM_PRCR = SYSTEM_PRCR;                                    // Backup state of unlock register
  SYSTEM_PRCR  = ((0xA5 << 8) | (1 << 0) | (1 << 1) | (1 << 2)); // Unlock configuration registers.
  SYSTEM_VOCR  = 0;                                              // Enable input pins. Without this, the QSPI controller would be unable to read anything from input pins
  SYSTEM_PRCR  = ((0xA5 << 8) | rSYSTEM_PRCR);                   // Restore state of unlock register again
}

/*********************************************************************
*
*       _InitClocks()
*
*  Function description
*   Initiates the QSPI controller.
*/
static void _InitQSPI(void) {
  MSTP_MSTPCRB &= (~(1UL << 6UL));  // QSPI controller is disabled by default. Enable it.
}

/*********************************************************************
*
*       _InitPins()
*
*  Function description
*   Initiates the pins that are connected to the QSPI flash on this board.
*/
static void _InitPins(void) {
  U8 rPMISC_PWPR;
  //
  // Unlock pin config registers
  //
  rPMISC_PWPR = PMISC_PWPR; // Backup current state
  PMISC_PWPR  = (0 << 7);
  PMISC_PWPR  = (1 << 6);
  //
  // Configure pins for QSPI single mode
  //              Direction (0: input, 1: output)  Mode (GPIO or peripheral)   Function (here QSPI)    Standard driving capability   
  PFS_QSPCLK_A =  (1 << 2)                         | (1 << 16)                 | (0x11 << 24)          | (1 << 11);
  PFS_QSSL_A   =  (1 << 2)                         | (1 << 16)                 | (0x11 << 24)          | (1 << 11);
  PFS_QIO0_A   =  (1 << 2)                         | (1 << 16)                 | (0x11 << 24)          | (1 << 11);
  PFS_QIO1_A   =  (0 << 2)                         | (1 << 16)                 | (0x11 << 24)          | (1 << 11);
  //
  // Restore unlock register again
  //
  PMISC_PWPR = (0 << 7);
  PMISC_PWPR = rPMISC_PWPR;
}

/*********************************************************************
*
*       _FuncInQSPI()
*
*  Function description
*   Example function located in the QSPI flash.
*   Prints something via RTT.
*/
static void _FuncInQSPI(void) {
  int i;

  for (i = 0; i < 100; ++i) {
    printf("Hello World %d!\n", i);
  }
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       main()
*
*  Function description
*   Application entry point.
*/
int main(void) {
  //
  // Initiate clocks, the pins necessary for QSPI and the QSPI controller itself.
  // After this procedure, the QSPI flash should be memory mapped at 0x60000000
  //
  _InitClocks();
  _InitPins();
  _InitQSPI();
  //
  // Call function that is located in QSPI flash
  //
  _FuncInQSPI();
  return 0;
}

/*************************** End of file ****************************/
